
;;
; This programme parses XML documents containing XLink elements
; It returns an SXML presentation for a document plus additional information
; extracted from XLink markup (described below)
;
; 'SSAX:XML->SXML+xlink' function is the core of the programme. This funcion
; is a modified Oleg Kiselyov's 'SSAX:XML->SXML' function.
; 'SSAX:XML->SXML+xlink' has a complicated seed which consists of ten elements:
;  xlink:seed = (list  mode  traversal-info  roles  arcroles  stack
;                      locators+resources  arcs
;                      declared-labels  linkbase-labels)
; 
; 1. mode = 'general, 'extended or 'none. They have the following meaning:
;  - 'general - there are no XLink elements among current element's ancestors.
; So, 'extended' or 'simple' elements are expected (others don't have any XLink
; semantical meaning)
;  - 'extended - for elements that are direct children of an extended link
; element, i.e. 'locator', 'resource' or 'arc'
;  - 'none - no XLink elements are expected niether in the current element
; nor in any of its descendants
;
; 2. traversal-info - contains information extracted from XLink elements. 
;  traversal-info = (list  one-to-many-traverse
;                          one-to-many-traverse
;                          ...)
;  one-to-many-traverse = 
;    = (list  uri
;             linkbase
;             fragment
;             role
;             position
;             element
;             single-traverse  single-traverse  ...  single-traverse)
;  uri - can be 1) a string representing a Unified Resource Identifier of a
; resource participating in a link, or 2) #f, if this resource resides in the
; same document as a currently processed one 
;  linkbase - a boolean value: whether this uri is an XLink linkbase
;  fragment - a fragment within a resource defined by the uri. Fragment can
; be 1) An SXML presentation of the element (when the resource is a local one,
; defined by a 'resource' or 'simple' XLink elements;  uri=#f in that case),
; 2) A string representing an XPointer-defined location, or 3) #f if the whole
; resource defined by the uri participates in a link
;  role - a string representing an xlink:role, or #f if this attribute was not
; supplied for that resource
;  position - a number. It is a position within the document where the
; declaration of this resource took place
;  element - an SXML presentation of an element where the declaration of this
; resource took place
;  single-traverse - a traversal that can be initiated from that resource.
; There can be no 'single-traverse's for a 'one-to-many-traverse' if
; no traversal can be initiated (i.e. no arc starts from that resource)
;
;  single-traverse = (list  uri  fragment  role  position  element
;                           arcrole  show  actuate)
;  uri, fragment, role, position, element - similar properties for an ending
; resource 
;  arcrole - a string representing a role of the are, or #f if an xlink:arcrole
; attribute wasn't supplied for this arc
;  show - a string representing the value of xlink:show attribute, or #f if
; this attribute was omitted 
;  actuate - the same for an xlink:actuate attribute
;
; 3. roles - roles encountered in a processed document
;  roles = (list  role-info
;                 role-info
;                 ...)
;  role-info = (list  role  position  element)
;
; 4. arcroles - arcroles encountered in a processed document
;  arcroles = (list  arcrole-info
;                    arcrole-info
;                    ...)
;  arcrole-info = (list  arcrole  position  element)
;
; 5. stack - a list of stack-elements. This list has the following semantics:
; - new stack-element is added when the beginning of each element is processed
; - the stack-element is consumed at the finish-element (of the same element)
;  stack = (list  stack-element
;                 stack-element
;                 ...)
;  stack-element = (list  position  xlink-values)
;  position - a position within a file
;  xlink-values = (list  type  href  role  arcrole  show  actuate  label  from  to)
; where, for example, 'type' is the value of xlink:type attribute or #f if 
; there is no such attribute
;
; The other parameters of the seed are presented when an extended link is
; processed
;
; 6. locators+resources - locator and resource elements defined within an
; extended link. They are temporarily stored in this parameter. This info
; is converted into an 'traversal-info' parameter when the end-tag for an
; extended link element is encountered
;  locators+resources = (list  locator-or-resource
;                              locator-or-resource
;                              ...)
;  locator-or-resource = (list  uri  fragment  role  label
;                               position  element)
;  label - a string representing the value of xlink:label attribute, or #f if
; this attribute was omitted 
;
; 7. arcs - information about arce defined within an extended link. This info
; is converted into an 'traversal-info' parameter when the end-tag for an
; extended link element is encountered
;  arcs = (list  arc-info
;                arc-info
;                ...)
;  arc-info = (list  arcrole  show  actuate  from  to
;                    position  element)
;  from - a string representing the value of xlink:from attribute, or #f if
; this attribute was omitted 
;  to - the same for an xlink:to attribute
;
; 8. declared-labels - labels declared within an extended link. This parameter
; is used for constraint checking
;  declared-labels = (list  label  label ...)
;  label - a string
;
; 9. linkbase-labels - labels which are linkbases
;  linkbase-labels = (list  label  label ...)
;                    or #f
; #f appears in case if all labels are linkbases (such a situation is possible)
;
;  ----------------
; 
;   THAT MUST BE REMOVED:
; The result of an 'SSAX:XML->SXML+xlink' function is a list of four elements:
;  (list  sxml-tree
;         traversal-info
;         roles
;         arcroles)
;  sxml-tree - an SXML presentation of a document. It is the final value of the
; original-seed. The result is equal to the result of SSAX:XML->SXML function,
; if applyed to the same document
;  traversal-info - see paragraph 3 above
;  roles - see paragraph 4 above
;  arcroles - see paragraph 5 above


; Some global constants
(define xlink:namespace-uri 'http://www.w3.org/1999/xlink)
(define xlink:linkbase-uri "http://www.w3.org/1999/xlink/properties/linkbase")



;=========================================================================
; A 'seed' datatype
;  xlink:seed = (list  mode  traversal-info  roles  arcroles  stack
;                      locators+resources  arcs
;                      declared-labels  linkbase-labels)
; The last four parameters are optional. See a head comment for details

;------------------------------------------------
; Two mutator functions for a seed
; They are introducted in order to control (possible) future modifications of
; a 'seed' list

; This function constructs a seed consisting only of six compulsory elements
(define (xlink:make-small-seed mode traversal-info roles arcroles stack)
  (list mode traversal-info roles arcroles stack))


; The similar function which makes a full-length seed
(define (xlink:make-full-seed mode traversal-info roles arcroles stack
                              locators+resources arcs
                              declared-labels linkbase-labels)
  (list mode traversal-info roles arcroles stack
        locators+resources arcs declared-labels  linkbase-labels))
  

;------------------------------------------------
; Accessor functions

(define (xlink:seed-mode seed)
  (car seed))

(define (xlink:seed-traversal-info seed)
  (cadr seed))

(define (xlink:seed-roles seed)
  (list-ref seed 2))

(define (xlink:seed-arcroles seed)
  (list-ref seed 3))

(define (xlink:seed-stack seed)
  (list-ref seed 4))

; We assume that the seed has the full length for the latter four functions
(define (xlink:seed-locators+resources seed)
  (list-ref seed 5))

(define (xlink:seed-arcs seed)
  (list-ref seed 6))

(define (xlink:seed-declared-labels seed)
  (list-ref seed 7))

(define (xlink:seed-linkbase-labels seed)
  (list-ref seed 8))



;=========================================================================
; Here basic functions for special datatypes are collected

;------------------------------------------------
; 2. 'traversal-info' datatype
; traversal-info - contains information extracted from XLink elements.
;  traversal-info = (list  one-to-many-traverse
;                          one-to-many-traverse
;                          ...)
;  one-to-many-traverse = 
;    = (list  uri
;             linkbase
;             fragment
;             role
;             position
;             element
;             single-traverse  single-traverse  ...  single-traverse)
;  single-traverse = (list  uri  fragment  role  position  element
;                           arcrole  show  actuate)
; See a head comment for details

; Trivial mutator functions
(define (xlink:make-single-traverse uri fragment role position element
                                    arcrole show actuate)
  (list uri fragment role position element arcrole show actuate))

(define (xlink:make-one-to-many-traverse uri linkbase fragment role
                                         position element single-traverses)
  (cons* uri linkbase fragment role position element single-traverses))


; Accessor functions for a 'one-to-many-traverse' datatype
(define (xlink:one-to-many-uri one-to-many-traverse)
  (car one-to-many-traverse))

(define (xlink:one-to-many-linkbase one-to-many-traverse)
  (cadr one-to-many-traverse))

(define (xlink:one-to-many-fragment one-to-many-traverse)
  (list-ref one-to-many-traverse 2))

(define (xlink:one-to-many-role one-to-many-traverse)
  (list-ref one-to-many-traverse 3))

(define (xlink:one-to-many-position one-to-many-traverse)
  (list-ref one-to-many-traverse 4))

(define (xlink:one-to-many-element one-to-many-traverse)
  (list-ref one-to-many-traverse 5))

(define (xlink:one-to-many-single-traverses one-to-many-traverse)
  (list-tail one-to-many-traverse 6))


; Accessor functions for a 'traversal' datatype
(define (xlink:single-traverse-uri single-traverse)
  (car single-traverse))

(define (xlink:single-traverse-fragment single-traverse)
  (cadr single-traverse))

(define (xlink:single-traverse-role single-traverse)
  (list-ref single-traverse 2))

(define (xlink:single-traverse-position single-traverse)
  (list-ref single-traverse 3))

(define (xlink:single-traverse-element single-traverse)
  (list-ref single-traverse 4))

(define (xlink:single-traverse-arcrole single-traverse)
  (list-ref single-traverse 5))

(define (xlink:single-traverse-show single-traverse)
  (list-ref single-traverse 6))

(define (xlink:single-traverse-actuate single-traverse)
  (list-ref single-traverse 7))


; This function adds simple link's information to 'traversal-info'
(define (xlink:add-simple xlink-values element position traversal-info)
  (let((href (xlink:values-href xlink-values))
       (role (xlink:values-role xlink-values))
       (arcrole (xlink:values-arcrole xlink-values))
       (show (xlink:values-show xlink-values))
       (actuate (xlink:values-actuate xlink-values)))
    (if(not href)   ; the link is untraversable
       (cons
        (xlink:make-one-to-many-traverse
         #f #f element #f position element '())
        traversal-info)
       (let((linkbase (equal? arcrole xlink:linkbase-uri))
            (lst (string-split href (list #\#) 2)))
         (let-values*
             (((uri fragment) 
               (cond
                 ((= (length lst) 1) (values (car lst) #f))
                 ((= (string-length (car lst)) 0) (values #f (cadr lst)))
                 (else (values (car lst) (cadr lst))))))
           (let((single-traverse
                 (xlink:make-single-traverse uri fragment role position
                                             element arcrole show actuate)))
             (cons*
              (xlink:make-one-to-many-traverse    ; remote resource
               uri linkbase fragment role position element '())
              (xlink:make-one-to-many-traverse        ; local resource
               #f #f element #f position element (list single-traverse))
              traversal-info)))))))


; This function appends information to 'traversal-info' according to
; 'locators+resources', 'arcs' and 'linkbase-labels' parameters.
; The function is called at the end-tag of an extended link element.
(define (xlink:add-extended locators+resources arcs linkbase-labels
                            traversal-info)
  (let loop ((loc+res locators+resources)
             (traversal-info traversal-info))
    (if
     (null? loc+res)
     traversal-info    
     (let((locator-or-resource (car loc+res)))
       (let((label (xlink:locator-or-resource-label locator-or-resource)))
         (let rpt ((arcs arcs)
                   (single-traverses '()))
           (if
            (null? arcs)              
            (let((uri
                  (xlink:locator-or-resource-uri locator-or-resource))
                 (fragment 
                  (xlink:locator-or-resource-fragment locator-or-resource))
                 (role
                  (xlink:locator-or-resource-role locator-or-resource))                 
                 (position
                  (xlink:locator-or-resource-position locator-or-resource))
                 (element
                  (xlink:locator-or-resource-element locator-or-resource)))
              (let((linkbase 
                    (xlink:within-linkbase-labels? label linkbase-labels)))
                (loop
                 (cdr loc+res)
                 (cons
                  (xlink:make-one-to-many-traverse
                   uri linkbase fragment role
                   position element single-traverses)
                  traversal-info))))
            (let((arc-info (car arcs)))
              (let((from (xlink:arc-info-from arc-info))
                   (to (xlink:arc-info-to arc-info))
                   (arcrole (xlink:arc-info-arcrole arc-info))
                   (show (xlink:arc-info-show arc-info))
                   (actuate (xlink:arc-info-actuate arc-info)))
                (if
                 (and from (not (equal? from label)))
                 (rpt (cdr arcs) single-traverses)
                 (let back ((ls locators+resources)
                            (sts single-traverses))
                   (if
                    (null? ls)
                    (rpt (cdr arcs) sts)
                    (let((l-or-r (car ls)))
                      (let((label2
                            (xlink:locator-or-resource-label l-or-r))
                           (uri2
                            (xlink:locator-or-resource-uri l-or-r))
                           (fragment2
                            (xlink:locator-or-resource-fragment l-or-r))
                           (role2
                            (xlink:locator-or-resource-role l-or-r))
                           (position2
                            (xlink:locator-or-resource-position l-or-r))
                           (element2
                            (xlink:locator-or-resource-element l-or-r)))
                        (if
                         (and to (not (equal? to label2)))
                         (back (cdr ls) sts)
                         (back 
                          (cdr ls)
                          (cons
                           (xlink:make-single-traverse 
                            uri2 fragment2 role2 position2 element2
                            arcrole show actuate)
                           sts)))))))))))))))))
                                                      
                           
;------------------------------------------------
; 3. 'roles' datatype
; roles - roles encountered in a processed document
;  roles = (list  role-info
;                 role-info
;                 ...)
;  role-info = (list  role  position  element)

; If an xlink:role attribute is presented in 'xlink-values', information about it
; is added to 'roles'. Otherwise, 'roles' remain unchainged
(define (xlink:add-role xlink-values position element roles)
  (let((role (xlink:values-role xlink-values)))
    (if(not role)
       roles
       (cons (list role position element) roles))))
        

;------------------------------------------------
; 4. 'arcrole' datatype
; arcroles - arcroles encountered in a processed document
;  arcroles = (list  arcrole-info
;                    arcrole-info
;                    ...)
;  arcrole-info = (list  arcrole  position  element)

; If an xlink:arcrole attribute is presented in 'xlink-values', information about it
; is added to 'arcroles'. Otherwise, 'arcroles' remain unchainged
(define (xlink:add-arcrole xlink-values position element arcroles)
  (let((arcrole (xlink:values-arcrole xlink-values)))
    (if(or (not arcrole) (equal? arcrole xlink:linkbase-uri))
       arcroles
       (cons (list arcrole position element) arcroles))))


;------------------------------------------------
; 6. 'locators+resources' datatype
; locators+resources - locator and resource elements defined within an
; extended link. They are temporarily stored in this parameter. This info
; is converted into an 'traversal-info' parameter when the end-tag for an
; extended link element is encountered
;  locators+resources = (list  locator-or-resource
;                              locator-or-resource
;                              ...)
;  locator-or-resource = (list  uri  fragment  role  label
;                               position  element)

; Trivial accessor functions
(define (xlink:locator-or-resource-uri locator-or-resource)
  (car locator-or-resource))

(define (xlink:locator-or-resource-fragment locator-or-resource)
  (cadr locator-or-resource))

(define (xlink:locator-or-resource-role locator-or-resource)
  (list-ref locator-or-resource 2))

(define (xlink:locator-or-resource-label locator-or-resource)
  (list-ref locator-or-resource 3))

(define (xlink:locator-or-resource-position locator-or-resource)
  (list-ref locator-or-resource 4))

(define (xlink:locator-or-resource-element locator-or-resource)
  (list-ref locator-or-resource 5))


; If the following XLink constraint is fulfilled, this function adds
; information about the XLink locator element to 'locators+resources'.
; Otherwise, the funtion displays an error message and doesn't add anything.
;  Constraint: Attributes on Locator Element
;  The locator-type element must have the locator attribute (see 5.4 Locator
;  Attribute (href)). The locator attribute (href) must have a value supplied.
(define (xlink:add-locator xlink-values position element locators+resources)
  (let((href (xlink:values-href xlink-values))
       (role (xlink:values-role xlink-values))
       (label (xlink:values-label xlink-values)))
    (cond
      ((not href)
       (xlink:error position 
                    "locator element doesn't have an xlink:href attribute")
       locators+resources)
      (else
       (let((lst (string-split href (list #\#) 2)))
         (let-values*
             (((uri fragment) 
               (cond
                 ((= (length lst) 1) (values (car lst) #f))
                 ((= (string-length (car lst)) 0) (values #f (cadr lst)))
                 (else (values (car lst) (cadr lst))))))
              (cons
               (list uri fragment role label position element)
               locators+resources)))))))


; This function adds information concerning XLink resource element to
; 'locators+resources'
(define (xlink:add-resource xlink-values position element locators+resources)
  (let((role (xlink:values-role xlink-values))
       (label (xlink:values-label xlink-values)))
    (cons (list #f element role label position element) locators+resources)))


;------------------------------------------------
; 7. 'arcs' datatype
; arcs - information about arce defined within an extended link. This info
; is converted into an 'traversal-info' parameter when the end-tag for an
; extended link element is encountered
;  arcs = (list  arc-info
;                arc-info
;                ...)
;  arc-info = (list  arcrole  show  actuate  from  to
;                    position  element)

; Trivial accessor functions
(define (xlink:arc-info-arcrole arc-info)
  (car arc-info))

(define (xlink:arc-info-show arc-info)
  (cadr arc-info))

(define (xlink:arc-info-actuate arc-info)
  (list-ref arc-info 2))

(define (xlink:arc-info-from arc-info)
  (list-ref arc-info 3))

(define (xlink:arc-info-to arc-info)
  (list-ref arc-info 4))

(define (xlink:arc-info-position arc-info)
  (list-ref arc-info 5))

(define (xlink:arc-info-element arc-info)
  (list-ref arc-info 6))


; This function adds arc information to 'arcs'
; A side effect - the function checks the following XLink constraint:
;  Constraint: No Arc Duplication
;  Each arc-type element must have a pair of from and to xlink-values that does not
;  repeat the from and to xlink-values (respectively) for any other arc-type element
;  in the same extended link; that is, each pair in a link must be unique.
(define (xlink:add-arc xlink-values position element arcs)
  (let((arcrole (xlink:values-arcrole xlink-values))
       (show (xlink:values-show xlink-values))
       (actuate (xlink:values-actuate xlink-values))
       (from (xlink:values-from xlink-values))
       (to (xlink:values-to xlink-values)))
    (let loop ((as arcs))
      (if
       (null? as)
       (cons (list arcrole show actuate from to position element) arcs)
       (let((from2 (xlink:arc-info-from (car as)))
            (to2 (xlink:arc-info-to (car as))))
         (cond
           ((and (or (not from) (not from2) (equal? from from2))
                 (or (not to) (not to2) (equal? to to2)))
            (xlink:error position "duplicate arcs - xlink:from"
                         (if from (string-append "=" from) " - omitted")
                         ", xlink:to"
                         (if to (string-append "=" to) " - omitted"))
            (cons (list arcrole show actuate from to position element) arcs))
           (else
            (loop (cdr as)))))))))


; XLink specification, 5.1.3:
;  If no arc-type elements are provided in an extended link, then by extension
;  the missing from and to xlink-values are interpreted as standing for all the
;  labels in that link.
; This funtion inserts such a default arc if 'arcs' are empty
(define (xlink:add-default-arc arcs)
  (if(null? arcs)
     (list (list #f #f #f #f #f 0 '()))
     arcs))


;------------------------------------------------
; 8. 'declared-labels' datatype
; declared-labels - labels declared within an extended link. This parameter
; is used for constraint checking
;  declared-labels = (list  label  label ...)
;  label - a string

; If an xlink:label attribute is presented in 'xlink-values', it's value is added
; to 'declared-labels'. Otherwise, 'declared-labels' remain unchainged
(define (xlink:add-declared-label xlink-values declared-labels)
  (let((label (xlink:values-label xlink-values)))
    (if(not label)
       declared-labels
       (cons label declared-labels))))


; The function checks the following XLink constraint
;  Constraint: label, from, and to xlink-values
;  The value of a label, from, or to attribute must be an NCName. If a value 
;  is supplied for a from or to attribute, it must correspond to the same value
;  for some label attribute on a locator- or resource-type element that appears
;  as a direct child inside the same extended-type element as does the arc-type 
;  element. 
; Error message is displayed if some label was undeclared.
; The function always returns #t.
; It is called at the end-tag of an extended link element
(define (xlink:all-labels-declared arcs declared-labels)
  (let loop ((arcs arcs))
    (if
     (null? arcs)
     #t
     (let((arc-info (car arcs)))
       (let((from (xlink:arc-info-from arc-info))
            (to (xlink:arc-info-to arc-info))
            (position (xlink:arc-info-position arc-info)))       
         (when (and from (not (member from declared-labels)))
           (xlink:error position "label not defined - xlink:from=" from))
         (when (and to (not (member to declared-labels)))
           (xlink:error position "label not defined - xlink:to=" to))
         (loop (cdr arcs)))))))


;------------------------------------------------
; 9. 'linkbase-labels' datatype
; linkbase-labels - labels which are linkbases
;  linkbase-labels = (list  label  label  ...)
;                 or #f
; #f appears in case if all labels are linkbases (such a situation is possible)

; If an xlink:arcrole attribute is presented in 'xlink-values' and it is a
; linkbase uri, xlink:to attribute's value is added to 'linkbase-labels'.
; Otherwise, 'linkbase' labels remain unchanged.
(define (xlink:add-linkbase-label xlink-values linkbase-labels)
  (let((arcrole (xlink:values-arcrole xlink-values))
       (label (xlink:values-to xlink-values)))
    (if(equal? arcrole xlink:linkbase-uri)  
       (cond((not label) #f)
            ((equal? linkbase-labels #f) linkbase-labels)
            ((member label linkbase-labels) linkbase-labels)
            (else (cons label linkbase-labels)))
       linkbase-labels)))


; This predicate checks whether a label is a linkbase. Again, the function is
; less trivial than 'member' because of special cases
(define (xlink:within-linkbase-labels? label linkbase-labels)
  (cond((equal? linkbase-labels #f) #t)
       ((equal? label #f) #f)
       ((member label linkbase-labels) #t)
       (else #f)))



;=========================================================================
; Some simple functions working with attributes

;------------------------------------------------
; Trivial accessor and mutator functions
; These functions are used as a level of abstraction

; The function constructs a datatype (just a list in a current implementation)
; which contains xlink-values of all xlink-related attributes. For example, 'type' is
; the value of xlink:type attribute or #f if there is no such attribute.
; This datatype will be called 'xlink-values' in the latter text
(define (xlink:construct-xlink-values type href role arcrole show
                                actuate label from to)
  (list type href role arcrole show actuate label from to))


; The next 9 functions are used for getting an attribute value from 'xlink-values'
; datatype
;  xlink-values = (type href role arcrole show actuate label from to)
(define (xlink:values-type xlink-values)
  (car xlink-values))

(define (xlink:values-href xlink-values)
  (cadr xlink-values))

(define (xlink:values-role xlink-values)
  (list-ref xlink-values 2))

(define (xlink:values-arcrole xlink-values)
  (list-ref xlink-values 3))

(define (xlink:values-show xlink-values)
  (list-ref xlink-values 4))

(define (xlink:values-actuate xlink-values)
  (list-ref xlink-values 5))

(define (xlink:values-label xlink-values)
  (list-ref xlink-values 6))

(define (xlink:values-from xlink-values)
  (list-ref xlink-values 7))

(define (xlink:values-to xlink-values)
  (list-ref xlink-values 8))


;------------------------------------------------
; Functions which reads attributes

; The function is given a list called 'attributes' (in SSAX parser). This list
; has the form
;  attributes = (list  attribute  
;                      attribute
;                      ...)
;  attribute = (cons  (cons  namespace-prefix  attribute-name)
;                     attribute-value )
;           or (cons  attribute-name  attribute-value )
;  namespaces - defined in "ssax.scm"
; The function reads XLink attributes' values and returns a 'xlink-values'
; datatype (the result of 'xlink:construct-xlink-values' function)
(define (xlink:read-attributes attributes namespaces)
  (let loop ((attributes attributes)
             (type #f) (href #f) (role #f) (arcrole #f) (show #f) (actuate #f)
             (label #f) (from #f) (to #f))
    (if(null? attributes)  ; the attribute list is over
       (xlink:construct-xlink-values type href role arcrole show actuate 
                               label from to)
       (let((attribute (car attributes)))                  
         (if(not (pair? (car attribute))) ;attribute doesn't have any namespace
            (loop (cdr attributes) 
                  type href role arcrole show actuate label from to)
            (let((namespace-prefix (caar attribute))
                 (attribute-name (cdar attribute))
                 (attribute-value (cdr attribute)))
              (let((namespace-uri
                    (let rpt ((ns namespaces))
                      (cond
                        ((null? ns) namespace-prefix)
                        ((equal? (cadar ns) namespace-prefix) (cddar ns))
                        (else (rpt (cdr ns)))))))
              (if(not (equal? namespace-uri xlink:namespace-uri))
                 (loop (cdr attributes) 
                       type href role arcrole show actuate label from to)
                 (case attribute-name
                   ((type) (loop (cdr attributes) attribute-value href role 
	                         arcrole show actuate label from to))
                   ((href) (loop (cdr attributes) type attribute-value role
			         arcrole show actuate label from to))
                   ((role) (loop (cdr attributes) type href attribute-value 
                                 arcrole show actuate label from to))
                   ((arcrole) (loop (cdr attributes) type href role 
                                 attribute-value show actuate label from to))
                   ((show) (loop (cdr attributes) type href role arcrole 
                                 attribute-value actuate label from to))
                   ((actuate) (loop (cdr attributes) type href role arcrole 
                                    show attribute-value label from to))
                   ((label) (loop (cdr attributes) type href role arcrole 
                                  show actuate attribute-value from to))
                   ((from) (loop (cdr attributes) type href role arcrole 
                                 show actuate label attribute-value to))
                   ((to) (loop (cdr attributes) type href role arcrole show 
                               actuate label from attribute-value))
                   (else (loop (cdr attributes) type href role arcrole show 
                               actuate label from to)))))))))))


; The function reads 'element's attributes
;  element - an SXML node representing an element
;  ns-prefixes = (list  (list  prefix  namespace-uri)
;                       (list  prefix  namespace-uri)
;                       ...)
;  prefix - a symbol
;  namespace-uri - a string
; An 'xlink-values' datatype is returned
(define (xlink:read-SXML-attributes element ns-prefixes)
  (let((attr-node ((select-kids (ntype?? '@)) element)))
    (if
     (null? attr-node)  ; no attributes
     (xlink:construct-xlink-values #f #f #f #f #f #f #f #f #f)
     (let loop ((attr-list (cdar attr-node))
                (type #f) (href #f) (role #f) (arcrole #f) (show #f)
                (actuate #f) (label #f) (from #f) (to #f))
       (if
        (null? attr-list)
        (xlink:construct-xlink-values
         type href role arcrole show actuate label from to)
        (let((attribute-name (symbol->string (caar attr-list)))
             (attribute-value (cadar attr-list)))
          (let-values*
              (((prefix local)
                (cond
                  ((string-rindex attribute-name #\:)
                   => (lambda (pos)
                        (values
                         (string->symbol (substring attribute-name 0 pos))
                         (string->symbol
                          (substring attribute-name (+ pos 1)
                                     (string-length attribute-name))))))
                  (else
                   (values #f attribute-name)))))
            (if
             (not prefix)   ; this is a non-qualified name
             (loop (cdr attr-list)
                   type href role arcrole show actuate label from to)
             (let((namespace-uri
                   (cond
                     ((assoc prefix ns-prefixes)
                      => (lambda (pair)
                           (string->symbol (cadr pair))))
                     (else
                      prefix))))
               (if
                (not (equal? namespace-uri xlink:namespace-uri))
                (loop (cdr attr-list)
                   type href role arcrole show actuate label from to)
                (case local
                  ((type) (loop (cdr attr-list) attribute-value href role 
                                arcrole show actuate label from to))
                  ((href) (loop (cdr attr-list) type attribute-value role
                                arcrole show actuate label from to))
                  ((role) (loop (cdr attr-list) type href attribute-value 
                                arcrole show actuate label from to))
                  ((arcrole) (loop (cdr attr-list) type href role 
                                   attribute-value show actuate label from to))
                  ((show) (loop (cdr attr-list) type href role arcrole 
                                attribute-value actuate label from to))
                  ((actuate) (loop (cdr attr-list) type href role arcrole 
                                    show attribute-value label from to))
                  ((label) (loop (cdr attr-list) type href role arcrole 
                                 show actuate attribute-value from to))
                  ((from) (loop (cdr attr-list) type href role arcrole 
                                show actuate label attribute-value to))
                  ((to) (loop (cdr attr-list) type href role arcrole show 
                               actuate label from attribute-value))
                  (else (loop (cdr attr-list) type href role arcrole show 
                              actuate label from to)))))))))))))


;------------------------------------------------
; These functions check XLink constrains which limit some attributes' xlink-values

; A helper function which is used by the next one
;  value - a value of an attribute (#f if there is no such attribute)
;  valid-xlink-values - a list of xlink-values which are allowed for this attribute
;  attr-name - a string denotating a name of an attribute (for a message)
;  position - position within a file
; Function always returns #t. 
; Side effects: function "cerr"s a message if 'value' is not #f and not within
; 'valid-xlink-values'
(define (xlink:check-helper value valid-xlink-values attr-name position)
  (cond
    ((not value) )  ; a value is #f - a correct situation
    ((not (member value valid-xlink-values))
     (xlink:error position "unexpected attribute value - " 
                  attr-name "=" value))
    (else #t)))


;  xlink-values = (type href role arcrole show actuate label from to)
; where, for example, 'type' is the value of xlink:type attribute or #f if 
; there is no such attribute (this datatype is a result 
; of 'read-xlink-attributes' function)
;  position - position within a file
;
; The function checks the three similar XLink constraints:
;  1. Constraint: type Value
;  The value of the type attribute must be supplied. The value must be one of
;  "simple", "extended", "locator", "arc", "resource", "title", or "none".
;  2. Constraint: show Value
;  If a value is supplied for a show attribute, it must be one of the xlink-values
;  "new", "replace", "embed", "other", and "none".
;  3. Constraint: actuate Value
;  If a value is supplied for an actuate attribute, it must be be one of the 
;  xlink-values "onLoad", "onRequest", "other", and "none".
; 
; The result is always #t
; Side effects - error messages (printed by
; an 'xlink:check-helper' function above)
(define (xlink:check-type-show-actuate-constraints xlink-values position)
  (xlink:check-helper (xlink:values-type xlink-values)
                      '("simple" "extended" "locator" "arc" "resource"
                        "title" "none")
                      "xlink:type"
                      position)
  (xlink:check-helper (xlink:values-show xlink-values)
                      '("new" "replace" "embed" "other" "none")
                      "xlink:show"
                      position)
  (xlink:check-helper (xlink:values-actuate xlink-values)
                      '("onLoad" "onRequest" "other" "none")
                      "xlink:actuate"
                      position))
    


;=========================================================================
; Functions which perform starting and ending actions for XLink elements
; All these functions have the same signature:
;
; (smth-start position xlink-values xlink:seed)
;  position - position within a file
;  xlink-values = (list  type  href  role  arcrole  show  actuate  label  from  to)
; where, for example, 'type' is the value of xlink:type attribute or #f if 
; there is no such attribute
;  xlink:seed = (list  mode  traversal-info  roles  arcroles  stack
;                      locators+resources  arcs
;                      declared-labels  linkbase-labels)
; See a head comment for details
;
; (smth-end xlink:parent-seed xlink:seed element)
; element - the SXML presentation of the current element
;
; All the functions return a new 'xlink:seed'

;------------------------------------------------
; A general element
; It is the element which doesn't have any XLink meaning, but its descendants
; might have such a meaning

(define (xlink:general-start position xlink-values seed)
  (let((traversal-info (xlink:seed-traversal-info seed))
       (roles (xlink:seed-roles seed))
       (arcroles (xlink:seed-arcroles seed))
       (stack (cons (list position xlink-values) (xlink:seed-stack seed))))
    (xlink:make-small-seed 'general traversal-info roles arcroles stack)))


(define (xlink:general-end parent-seed seed element)
  (let((mode (xlink:seed-mode parent-seed))
       (traversal-info (xlink:seed-traversal-info seed))
       (roles (xlink:seed-roles seed))
       (arcroles (xlink:seed-arcroles seed))
       (stack (xlink:seed-stack parent-seed)))
    (xlink:make-small-seed mode traversal-info roles arcroles stack)))


;------------------------------------------------
; An element and all its descendants don't have any XLink meaning

(define (xlink:none-start position xlink-values seed)
  (let((stack (cons (list position xlink-values) (xlink:seed-stack seed))))
    (xlink:make-small-seed 'none '() '() '() stack)))


(define (xlink:none-end parent-seed seed element)
  parent-seed)

  
;------------------------------------------------
; A simple-link element

(define (xlink:simple-start position xlink-values seed)
  (let((stack (cons (list position xlink-values) (xlink:seed-stack seed))))
    (xlink:make-small-seed 'none '() '() '() stack)))


(define (xlink:simple-end parent-seed seed element)
  (let((stack-element (car (xlink:seed-stack seed))))
    (let((position (car stack-element))
         (xlink-values (cadr stack-element)))
      (let((mode (xlink:seed-mode parent-seed))
           (traversal-info (xlink:add-simple 
                            xlink-values element position
                            (xlink:seed-traversal-info parent-seed)))
           (roles (xlink:add-role xlink-values position element 
                                  (xlink:seed-roles parent-seed)))
           (arcroles (xlink:add-arcrole xlink-values position element 
                                        (xlink:seed-arcroles parent-seed)))
           (stack (xlink:seed-stack parent-seed)))
        (xlink:make-small-seed mode traversal-info roles arcroles stack)))))
          
                    
;------------------------------------------------
; An extended-link element

(define (xlink:extended-start position xlink-values seed)
  (let((traversal-info (xlink:seed-traversal-info seed))
       (arcroles (xlink:seed-arcroles seed))
       (stack (cons (list position xlink-values) (xlink:seed-stack seed))))
    (xlink:make-full-seed 'extended traversal-info '() arcroles stack
                          '() '() '() '())))


(define (xlink:extended-end parent-seed seed element)
  (let((stack-element (car (xlink:seed-stack seed))))
    (let((position (car stack-element))
         (xlink-values (cadr stack-element)))
      (let((locators+resources (xlink:seed-locators+resources seed))
           (arcs (xlink:add-default-arc (xlink:seed-arcs seed)))
           (declared-labels (xlink:seed-declared-labels seed))
           (linkbase-labels (xlink:seed-linkbase-labels seed)))      
        (xlink:all-labels-declared arcs declared-labels)      
      (let((mode (xlink:seed-mode parent-seed))
           (traversal-info
            (xlink:add-extended locators+resources arcs linkbase-labels
                                (xlink:seed-traversal-info seed)))
           (roles (append
                   (xlink:seed-roles seed)
                   (xlink:add-role xlink-values position element
                                   (xlink:seed-roles parent-seed))))
           (arcroles (xlink:add-arcrole xlink-values position element 
                                        (xlink:seed-arcroles seed)))
           (stack (xlink:seed-stack parent-seed)))
      (xlink:make-small-seed mode traversal-info roles arcroles stack))))))


;------------------------------------------------
; A locator element

(define (xlink:locator-start position xlink-values seed)
  (let((stack (cons (list position xlink-values) (xlink:seed-stack seed))))
    (xlink:make-small-seed 'none '() '() '() stack)))
  

(define (xlink:locator-end parent-seed seed element)
  (let((stack-element (car (xlink:seed-stack seed))))
    (let((position (car stack-element))
         (xlink-values (cadr stack-element)))
      (let((mode (xlink:seed-mode parent-seed))
           (traversal-info (xlink:seed-traversal-info parent-seed))
           (roles (xlink:add-role xlink-values position element
                                  (xlink:seed-roles parent-seed)))
           (arcroles (xlink:seed-arcroles parent-seed))
           (stack (xlink:seed-stack parent-seed))
           (locators+resources 
            (xlink:add-locator xlink-values position element
                               (xlink:seed-locators+resources parent-seed)))
           (arcs (xlink:seed-arcs parent-seed))
           (declared-labels 
            (xlink:add-declared-label 
             xlink-values (xlink:seed-declared-labels parent-seed)))
           (linkbase-labels (xlink:seed-linkbase-labels parent-seed)))
        (xlink:make-full-seed mode traversal-info roles arcroles stack
                              locators+resources arcs
                              declared-labels linkbase-labels)))))            


;------------------------------------------------
; A resource element

(define (xlink:resource-start position xlink-values seed)
  (let((stack (cons (list position xlink-values) (xlink:seed-stack seed))))
    (xlink:make-small-seed 'none '() '() '() stack)))


(define (xlink:resource-end parent-seed seed element)
  (let((stack-element (car (xlink:seed-stack seed))))
    (let((position (car stack-element))
         (xlink-values (cadr stack-element)))
      (let((mode (xlink:seed-mode parent-seed))
           (traversal-info (xlink:seed-traversal-info parent-seed))
           (roles (xlink:add-role xlink-values position element
                                  (xlink:seed-roles parent-seed)))
           (arcroles (xlink:seed-arcroles parent-seed))
           (stack (xlink:seed-stack parent-seed))
           (locators+resources 
            (xlink:add-resource xlink-values position element
                                (xlink:seed-locators+resources parent-seed)))
           (arcs (xlink:seed-arcs parent-seed))
           (declared-labels
            (xlink:add-declared-label
             xlink-values (xlink:seed-declared-labels parent-seed)))
           (linkbase-labels (xlink:seed-linkbase-labels parent-seed)))
        (xlink:make-full-seed mode traversal-info roles arcroles stack
                              locators+resources arcs
                              declared-labels linkbase-labels)))))
                               

;------------------------------------------------
; An arc element

(define (xlink:arc-start position xlink-values seed)
  (let((stack (cons (list position xlink-values) (xlink:seed-stack seed))))
    (xlink:make-small-seed 'none '() '() '() stack)))


(define (xlink:arc-end parent-seed seed element)
  (let((stack-element (car (xlink:seed-stack seed))))
    (let((position (car stack-element))
         (xlink-values (cadr stack-element)))
      (let((mode (xlink:seed-mode parent-seed))
           (traversal-info (xlink:seed-traversal-info parent-seed))
           (roles (xlink:seed-roles parent-seed))
           (arcroles (xlink:add-arcrole xlink-values position element
                                        (xlink:seed-arcroles parent-seed)))
           (stack (xlink:seed-stack parent-seed))
           (locators+resources (xlink:seed-locators+resources parent-seed))           
           (arcs (xlink:add-arc xlink-values position element
                                (xlink:seed-arcs parent-seed)))
           (declared-labels                
            (xlink:seed-declared-labels parent-seed))
           (linkbase-labels
            (xlink:add-linkbase-label
             xlink-values (xlink:seed-linkbase-labels parent-seed))))
        (xlink:make-full-seed mode traversal-info roles arcroles stack
                              locators+resources arcs
                              declared-labels linkbase-labels)))))



;=========================================================================
; Core features of the parser

; This function displays an error message. #t is returned
;  position - position within a file
;  text - a message to display
(define (xlink:error position . text)
  (cond-expand
   ((or bigloo  plt)
    (apply cerr (append (list nl "XLink error in position " position ":" nl)
                        text 
                        (list nl))))
   ((or gambit guile chicken)
    (apply cerr (append (list nl "XLink error in line " position ":" nl)
                        text 
                        (list nl))))
   (else 
    (apply cerr (append (list nl "XLink error:" nl)
                        text 
                        (list nl)))))
  #t)


;------------------------------------------------
; Handler units
              
; This function is called by the NEW-LEVEL-SEED handler
; A new 'xlink:seed' is returned
(define (xlink:new-level-seed-handler port attributes namespaces seed)
  (let((position 
        (cond-expand         
	 (bigloo (input-port-position port))
         (chicken (receive (row col) (port-position port) row))
         (gambit (port-input-line-count port))
         (guile (port-line port))
         (plt (file-position port))
         (else 13)))
       (xlink-values (xlink:read-attributes attributes namespaces)))
    (xlink:check-type-show-actuate-constraints xlink-values position)
    (let((mode (xlink:seed-mode seed))
         (type (xlink:values-type xlink-values)))
      (case mode
        ((general)
         (case (if type (string->symbol type) type)
           ((simple) (xlink:simple-start position xlink-values seed))
           ((extended) (xlink:extended-start position xlink-values seed))
           ((none) (xlink:none-start position xlink-values seed))
           (else (xlink:general-start position xlink-values seed))))
        ((extended)
         (case (if type (string->symbol type) type)
           ((locator) (xlink:locator-start position xlink-values seed))
           ((resource) (xlink:resource-start position xlink-values seed))
           ((arc) (xlink:arc-start position xlink-values seed))
           (else (xlink:none-start position xlink-values seed))))
        ((none) (xlink:none-start position xlink-values seed))
        (else 
         (xlink:error position "internal processor error - mode=" 
                      mode)
         (xlink:none-start position xlink-values seed))))))
  
  
; This function is called by the FINISH-ELEMENT handler
; A new 'xlink:seed' is returned
(define (xlink:finish-element-handler parent-seed seed element)
  (let((xlink-values (cadar (xlink:seed-stack seed))))
    (let((mode (xlink:seed-mode parent-seed))
         (type (xlink:values-type xlink-values)))
      (case mode
        ((general)
         (case (if type (string->symbol type) type)
           ((simple) (xlink:simple-end parent-seed seed element))
           ((extended) (xlink:extended-end parent-seed 
                                           seed element))
           ((none) (xlink:none-end parent-seed seed element))
           (else (xlink:general-end parent-seed seed element))))
        ((extended)
         (case (if type (string->symbol type) type)
           ((locator) (xlink:locator-end parent-seed 
                                         seed element))
           ((resource) (xlink:resource-end parent-seed 
                                           seed element))
           ((arc) (xlink:arc-end parent-seed seed element))
           (else (xlink:none-end parent-seed seed element))))
        ((none) (xlink:none-end parent-seed seed element))
        (else 
         (xlink:error 0 "internal processor error - mode=" 
                      mode)
         (xlink:none-end parent-seed seed element))))))


; This function constructs the member of an axuiliary list
(define (xlink:ending-action xlink:seed)
  (let((traversal-info (reverse (xlink:seed-traversal-info xlink:seed)))
       (roles (reverse (xlink:seed-roles xlink:seed)))
       (arcroles (reverse (xlink:seed-arcroles xlink:seed))))
    (list 'sxlink
          (cons 'traversal-info traversal-info)
          (cons 'roles roles)
          (cons 'arcroles arcroles))))


;------------------------------------------------
; The function which adds XLink-related information to the SXML document

;  document - an SXML document
; The function emulates a 'fold-ts' operation.
; A new SXML document is returned. It contains an auxiliary list with an
; 'sxlink' subtree. If the source document already contains such a
; subtree, it will be replaced. Other subtrees in an auxiliary list will
; remain unchanged.
(define (SXML->SXML+xlink document)
  
  (define (fold-ts node ns-prefixes seed)
    (let((xlink-values (xlink:read-SXML-attributes node ns-prefixes)))
      (let((mode (xlink:seed-mode seed))
           (type (xlink:values-type xlink-values)))
        (let rpt 
          ((kids ((select-kids (ntype?? '*)) node))
           (new-seed         
            (case mode
              ((general)
               (case (if type (string->symbol type) type)
                 ((simple) (xlink:simple-start 0 xlink-values seed))
                 ((extended) (xlink:extended-start 0 xlink-values seed))
                 ((none) (xlink:none-start 0 xlink-values seed))
                 (else (xlink:general-start 0 xlink-values seed))))                  
              ((extended)
               (case (if type (string->symbol type) type)
                 ((locator) (xlink:locator-start 0 xlink-values seed))
                 ((resource) (xlink:resource-start 0 xlink-values seed))
                 ((arc) (xlink:arc-start 0 xlink-values seed))
                 (else (xlink:none-start 0 xlink-values seed))))
              ((none) (xlink:none-start 0 xlink-values seed))
              (else 
               (xlink:error 0 "internal processor error - mode=" mode)
               (xlink:none-start 0 xlink-values seed)))))
          (if(not (null? kids))
             (rpt (cdr kids)
                  (fold-ts (car kids) ns-prefixes new-seed))
             (case mode
               ((general)
                (case (if type (string->symbol type) type)
                  ((simple) (xlink:simple-end seed new-seed node))
                  ((extended) (xlink:extended-end seed new-seed node))
                  ((none) (xlink:none-end seed new-seed node))
                  (else (xlink:general-end seed new-seed node))))
               ((extended)
                (case (if type (string->symbol type) type)
                  ((locator) (xlink:locator-end seed new-seed node))
                  ((resource) (xlink:resource-end seed new-seed node))
                  ((arc) (xlink:arc-end seed new-seed node))
                  (else (xlink:none-end seed new-seed node))))
               ((none) (xlink:none-end seed new-seed node))
               (else 
                (xlink:error 0 "internal processor error - mode=" mode)
                (xlink:none-end seed new-seed node))))))))
    
  (let((aux-subtrees
        (let((aux ((select-kids (ntype?? '@@)) document)))
          (if(null? aux)
             '()
             (let loop ((res '())
                        (to-see (cdar aux)))
               (cond
                 ((null? to-see) (reverse res))
                 ((equal? (caar to-see) 'sxlink) (loop res (cdr to-see)))
                 (else (loop (cons (car to-see) res)
                             (cdr to-see))))))))
       (ns-prefixes
        (let((ns-node ((select-kids (ntype?? '*NAMESPACES*)) document)))
          (if(null? ns-node)
             '()
             (cdar ns-node)))))
    (let((seed 
          (fold-ts document ns-prefixes
                   (xlink:make-small-seed 'general '() '() '() '()))))
      (let((traversal-info (xlink:seed-traversal-info seed))
           (roles (xlink:seed-roles seed))
           (arcroles (xlink:seed-arcroles seed))
           (kids ((select-kids
                   (lambda (node)
                     (not (and (pair? node) (equal? (car node) '@@)))))
                  document)))
        (cons* '*TOP*
               (cons* '@@
                      (list 'sxlink
                            (cons 'traversal-info traversal-info)
                            (cons 'roles roles)
                            (cons 'arcroles arcroles))
                      aux-subtrees)
               kids)))))
